# coding:utf-8
__author__ = 'dawei.leng'
__version__ = '1.48'

"""
 Another CTC implemented in theano.
 The `cost()` function of `CTC_Timescale` and `CTC_Logscale` classes return the average NLL over a batch samples given query sequences and score matrices.
 This implementation features:
    1) batch / mask supported.
    2) speed comparable with (~35% slower than) the numba implementation which is the fastest by now.

 Created   :  12, 10, 2015
 Revised   :   8,  3, 2016   ver 1.46
           :   1, 10, 2017   ver 1.47  fix wrong 'static_method' decorators.
           :   3, 20, 2017   ver 1.48  add support for alignment {'pre'/'post'} for `CTC_Logscale` class

 Reference :  [1] Alex Graves, etc., Connectionist temporal classification: labelling unsegmented sequence data with
                  recurrent neural networks, ICML, 2006
              [2] Alex Graves, Supervised sequence labelling with recurrent neural networks, 2014
              [3] Lawrence R. Rabiner, A tutorial on hidden Markov models and selected applications in speech recognition,
                  Proceedings of the IEEE, 1989
              [4] Maas Andrew, etc., https://github.com/amaas/stanford-ctc/blob/master/ctc_fast/ctc-loss/ctc_fast.pyx
              [5] Mohammad Pezeshki, https://github.com/mohammadpz/CTC-Connectionist-Temporal-Classification/blob/master/ctc_cost.py
              [6] Shawn Tan, https://github.com/shawntan/rnn-experiment/blob/master/CTC.ipynb
"""
import theano
from theano import tensor
from theano.ifelse import ifelse
floatX = theano.config.floatX

class CTC_Timescale(object):
    """
    Compute CTC cost using time normalization instead of log scale computation.
    Batch supported.
    To compute the batch cost, use .cost() function below.
    Speed slower than the numba & cython version (~6min vs ~3.9min on word_correction_CTC experiment), much faster than
    the following non-batch version ctc_path_probability().

    B: BATCH_SIZE
    L: query sequence length (maximum length of a batch)
    C: class number
    T: time length (maximum time length of a batch)
    """
    @classmethod
    def cost(self, queryseq, scorematrix, queryseq_mask=None, scorematrix_mask=None, blank_symbol=None, align='pre'):
        """
        Compute CTC cost, using only the forward pass
        :param queryseq: (L, B)
        :param scorematrix: (T, C+1, B)
        :param queryseq_mask: (L, B)
        :param scorematrix_mask: (T, B)
        :param blank_symbol: scalar, = C by default
        :return: negative log likelihood averaged over a batch
        """
        if blank_symbol is None:
            # blank_symbol = scorematrix.shape[1] - 1
            blank_symbol = tensor.cast(scorematrix.shape[1], floatX) - 1.0
        queryseq_padded, queryseq_mask_padded = self._pad_blanks(queryseq, blank_symbol, queryseq_mask)
        results = self.path_probability(queryseq_padded, scorematrix, queryseq_mask_padded, scorematrix_mask, blank_symbol, align)
        NLL = -results[1][-1]                                             # negative log likelihood
        NLL_avg = tensor.mean(NLL)                                        # batch averaged NLL, used as cost
        # if scorematrix_mask is not None:
        #     sm = scorematrix * scorematrix_mask.dimshuffle(0, 'x', 1)
        # else:
        #     sm = scorematrix
        # penalty = tensor.log(sm[:, blank_symbol, :].sum())
        return NLL_avg

    @classmethod
    def path_probability(self, queryseq_padded, scorematrix, queryseq_mask_padded=None, scorematrix_mask=None, blank_symbol=None):
        """
        Compute p(l|x) using only the forward variable
        :param queryseq_padded: (2L+1, B)
        :param scorematrix: (T, C+1, B)
        :param queryseq_mask_padded: (2L+1, B)
        :param scorematrix_mask: (T, B)
        :param blank_symbol: = C by default
        :return:
        """
        if blank_symbol is None:
            # blank_symbol = scorematrix.shape[1] - 1
            blank_symbol = tensor.cast(scorematrix.shape[1], floatX) - 1.0
        if queryseq_mask_padded is None:
                    queryseq_mask_padded = tensor.ones_like(queryseq_padded, dtype=floatX)

        pred_y = self._class_batch_to_labeling_batch(queryseq_padded, scorematrix, scorematrix_mask)  # (T, 2L+1, B), reshaped scorematrix

        r2, r3 = self._recurrence_relation(queryseq_padded, queryseq_mask_padded, blank_symbol)       # r2 (2L+1, 2L+1), r3 (2L+1, 2L+1, B)

        def step(p_curr, p_prev, LLForward, countdown, r2, r3, queryseq_mask_padded):
            """
            [DV, 1-14-2016]: A very weird problem encountered when integrating this CTC implementation into Keras. Before this revision
                             there were no input parameters (r2, r3, queryseq_mask_padded) specified, they just referred to the outer scope ones.
                             However, this will cause the CTC integrated within Keras producing inaccurate loss value, meanwhile when compiled
                             as a separate function, the returned ctc loss value is accurate anyway. But if with these 3 parameters added as
                             input, the problem vanished. This took me two days to find this remedy. I suspect this'd be the bug of theano.
            :param p_curr:     (2L+1, B), one column of scorematrix
            :param p_prev:     (B, 2L+1)
            :param LLForward:  (B, 1)
            :param countdown:  scalar
            :param r2:         (2L+1, 2L+1)
            :param r3:         (2L+1, 2L+1, B)
            :param queryseq_mask_padded: (2L+1, B)
            :return:
            """
            dotproduct = (p_prev + tensor.dot(p_prev, r2) +                                           # tensor.dot(p_prev, r2) = alpha(t-1, u-1)
                          (p_prev.dimshuffle(1, 'x', 0) * r3).sum(axis=0).T)                          # = alpha(t-1, u-2) conditionally
            p_curr = p_curr.T * dotproduct
            if queryseq_mask_padded is not None:
                p_curr *= queryseq_mask_padded.T                                                      # (B, 2L+1) * (B, 2L+1) * (B, 2L+1) = (B, 2L+1)
            start = tensor.max([0, queryseq_padded.shape[0] - 2 * countdown])
            mask = tensor.concatenate([tensor.zeros([queryseq_padded.shape[1], start]),
                                       tensor.ones([queryseq_padded.shape[1], queryseq_padded.shape[0] - start])], axis=1)
            p_curr *= mask
            c_batch = p_curr.sum(axis=1, keepdims=True)                                               # (B, 1)
            p_curr /= c_batch
            LLForward += tensor.log(c_batch)
            countdown -= 1
            return p_curr, LLForward, countdown                                                       # (B, 2L+1), (B, 1), scalar

        results, _ = theano.scan(
                step,
                sequences=[pred_y],                                                                   # scan only work on the first dimension
                outputs_info=[tensor.eye(queryseq_padded.shape[0])[0] * tensor.ones(queryseq_padded.T.shape),
                              tensor.unbroadcast(tensor.zeros([queryseq_padded.shape[1], 1]), 1), scorematrix.shape[0]],
                non_sequences=[r2, r3, queryseq_mask_padded])
        return results

    @classmethod
    def best_path_decode(self, scorematrix, scorematrix_mask=None, blank_symbol=None):
        """
        Computes the best path by simply choosing most likely label at each timestep
        :param scorematrix: (T, C+1, B)
        :param scorematrix_mask: (T, B)
        :param blank_symbol: = C by default
        :return: resultseq (T, B), resultseq_mask(T, B)
        Speed much slower than pure python version (normally ~40 times on HTR tasks)
        """
        bestlabels = tensor.argmax(scorematrix, axis=1)    # (T, B)
        T, Cp, B = scorematrix.shape
        resultseq, resultseq_mask = tensor.zeros([T, B], dtype=scorematrix.dtype)-1, tensor.zeros([T, B], dtype=scorematrix.dtype)
        if blank_symbol is None:
            # blank_symbol = Cp - 1.0
            blank_symbol = tensor.cast(Cp, floatX) - 1.0
        if scorematrix_mask is None:
            scorematrix_mask = tensor.ones([T, B], dtype=scorematrix.dtype)

        def step(labelseq, labelseq_mask, idx, resultseq, resultseq_mask, blank_symbol):
            seqlen = tensor.cast(labelseq_mask.sum(), 'int32')
            labelseq = self._remove_adjdup(labelseq[0:seqlen])
            labelseq = self._remove_value(labelseq, blank_symbol)
            seqlen2 = labelseq.size
            resultseq = tensor.set_subtensor(resultseq[0:seqlen2, idx], labelseq)
            resultseq_mask = tensor.set_subtensor(resultseq_mask[0:seqlen2, idx], tensor.ones_like(labelseq))
            idx += 1
            return idx, resultseq, resultseq_mask

        outputs, updates = theano.scan(fn = step,
                                       sequences=[bestlabels.T, scorematrix_mask.T],
                                       outputs_info=[0, resultseq, resultseq_mask],
                                       non_sequences=[blank_symbol],
                                       name='decode_scan')
        resultseq, resultseq_mask = outputs[1][-1], outputs[2][-1]
        return resultseq, resultseq_mask

    @classmethod
    def calc_CER(self, resultseq, targetseq, resultseq_mask=None, targetseq_mask=None):
        """
        Calculate the character error rate (CER) given ground truth 'targetseq' and CTC decoding output 'resultseq'
        :param resultseq (T1,  B)
        :param resultseq_mask (T1, B)
        :param targetseq (T2,  B)
        :param targetseq_mask (T2, B)
        :return: CER scalar
        """
        if resultseq_mask is None:
            resultseq_mask = tensor.ones_like(resultseq)
        if targetseq_mask is None:
            targetseq_mask = tensor.ones_like(targetseq)

        def step(result_seq, target_seq, result_seq_mask, target_seq_mask, TE, TG):
            L1 = tensor.cast(result_seq_mask.sum(), 'int32')
            L2 = tensor.cast(target_seq_mask.sum(), 'int32')
            d = self._editdist(result_seq[0:L1], target_seq[0:L2])
            TE += d
            TG += target_seq_mask.sum()
            return TE, TG

        outputs, updates = theano.scan(fn=step,
                                       sequences=[resultseq.T, targetseq.T, resultseq_mask.T, targetseq_mask.T],
                                       outputs_info=[tensor.zeros(1), tensor.zeros(1)],
                                       name='calc_CER')
        TE, TG = outputs[0][-1], outputs[1][-1]
        CER = TE/TG
        return CER, TE, TG

    @staticmethod
    def _remove_value(x, value):
        """
        Remove certain valued elements from a vector
        x: vector (must); value: scalar
        return a vector with all elements = 'value' removed
        """
        return (x - value).nonzero_values() + value

    @staticmethod
    def _remove_adjdup(x):
        """
        Remove adjacent duplicate items of a vector
        x: vector
        return a vector with adjacent duplicate items removed, for example [1,2,2,2,3,3,4] -> [1,2,3,4]
        """
        def update(x, nondup, idx):
            nondup = tensor.switch(tensor.eq(nondup[idx], x), nondup, tensor.set_subtensor(nondup[idx + 1], x))  # tensor.switch is much faster than ifelse
            idx = tensor.switch(tensor.eq(nondup[idx], x), idx, idx + 1)
            return nondup, idx
        nondup = x
        idx = tensor.as_tensor_variable(0)
        idx = tensor.cast(idx, 'int32')
        result, updates = theano.scan(fn = update, sequences=x, outputs_info=[nondup, idx], name='remove_adjdup')
        nondup = result[0][-1]
        idx = result[1][-1]
        return nondup[0:idx+1]

    @staticmethod
    def _editdist(s, t):
        """
        Levenshtein's edit distance function
        :param s: vector, source string
        :param t: vector, target string
        :return:  edit distance, scalar
        """
        def update(x, previous_row):
            current_row = previous_row + 1
            current_row = tensor.set_subtensor(current_row[1:], tensor.minimum(current_row[1:], tensor.add(previous_row[:-1], tensor.neq(target,x))))
            current_row = tensor.set_subtensor(current_row[1:], tensor.minimum(current_row[1:], current_row[0:-1] + 1))
            return current_row
        source, target = ifelse(tensor.lt(s.shape[0], t.shape[0]), (t, s), (s, t))
        previous_row = tensor.arange(target.size + 1, dtype=theano.config.floatX)
        result, updates = theano.scan(fn=update, sequences=source, outputs_info=previous_row, name='editdist')
        return result[-1,-1]

    @staticmethod
    def _pad_blanks(queryseq, blank_symbol, queryseq_mask=None):
        """
        Pad queryseq and corresponding queryseq_mask with blank symbol
        :param queryseq  (L, B)
        :param queryseq_mask (L, B)
        :param blank_symbol  scalar, must be float type!
        :return queryseq_padded, queryseq_mask_padded, both with shape (2L+1, B)
        """
        # for queryseq
        queryseq_extended = queryseq.dimshuffle(1, 0, 'x')                              # (L, B) -> (B, L, 1)
        blanks = tensor.zeros_like(queryseq_extended) + blank_symbol                    # (B, L, 1)
        concat = tensor.concatenate([queryseq_extended, blanks], axis=2)                # concat.shape = (B, L, 2)
        res = concat.reshape((concat.shape[0], concat.shape[1] * concat.shape[2])).T    # res.shape = (2L, B), the reshape will cause the last 2 dimensions interlace
        begining_blanks = tensor.zeros((1, res.shape[1])) + blank_symbol                # (1, B)
        queryseq_padded = tensor.concatenate([begining_blanks, res], axis=0)            # (1+2L, B)
        # for queryseq_mask
        if queryseq_mask is not None:
            queryseq_mask_extended = queryseq_mask.dimshuffle(1, 0, 'x')                          # (L, B) -> (B, L, 1)
            concat = tensor.concatenate([queryseq_mask_extended, queryseq_mask_extended], axis=2) # concat.shape = (B, L, 2)
            res = concat.reshape((concat.shape[0], concat.shape[1] * concat.shape[2])).T
            begining_blanks = tensor.ones((1, res.shape[1]), dtype=floatX)
            queryseq_mask_padded = tensor.concatenate([begining_blanks, res], axis=0)
        else:
            queryseq_mask_padded = None
        return queryseq_padded, queryseq_mask_padded

    @staticmethod
    def _class_batch_to_labeling_batch(queryseq_padded, scorematrix, scorematrix_mask=None):
        """
        Convert dimension 'class' of scorematrix to 'label'
        :param queryseq_padded: (2L+1, B)
        :param scorematrix: (T, C+1, B)
        :param scorematrix_mask: (T, B)
        :return: (T, 2L+1, B)
        """
        if scorematrix_mask is not None:
            scorematrix = scorematrix * scorematrix_mask.dimshuffle(0, 'x', 1)                   # (T, C+1, B) * (T, 1, B)
        batch_size = scorematrix.shape[2]  # = B
        res = scorematrix[:, queryseq_padded.astype('int32'), tensor.arange(batch_size)]         # (T, 2L+1, B), indexing each row of scorematrix with queryseq_padded
        return res

    @staticmethod
    def _recurrence_relation(queryseq_padded, queryseq_mask_padded=None, blank_symbol=None):
        """
        Generate structured matrix r2 & r3 for dynamic programming recurrence
        :param queryseq_padded: (2L+1, B)
        :param queryseq_mask_padded: (2L+1, B)
        :param blank_symbol: = C
        :return: r2 (2L+1, 2L+1), r3 (2L+1, 2L+1, B)
        """
        L2 = queryseq_padded.shape[0]                                                           # = 2L+1
        blanks = tensor.zeros((2, queryseq_padded.shape[1])) + blank_symbol                     # (2, B)
        ybb = tensor.concatenate((queryseq_padded, blanks), axis=0).T                           # (2L+3, B) -> (B, 2L+3)
        sec_diag = tensor.neq(ybb[:, :-2], ybb[:, 2:]) * tensor.eq(ybb[:, 1:-1], blank_symbol)  # (B, 2L+1)
        if queryseq_mask_padded is not None:
            sec_diag *= queryseq_mask_padded.T
        r2 = tensor.eye(L2, k=1)                                                                # upper diagonal matrix (2L+1, 2L+1)
        r3 = tensor.eye(L2, k=2).dimshuffle(0, 1, 'x') * sec_diag.dimshuffle(1, 'x', 0)         # (2L+1, 2L+1, B)
        return r2, r3

class CTC_Logscale(CTC_Timescale):
    """
    This implementation uses log scale computation.
    Batch supported. Note the log scale computation used to produce imprecise CTC cost in Theano (path probability).
    [Credits to Mohammad Pezeshki, https://github.com/mohammadpz/CTC-Connectionist-Temporal-Classification]
    B: BATCH_SIZE
    L: query sequence length (maximum length of a batch)
    C: class number
    T: time length (maximum time length of a batch)
    """
    @classmethod
    def cost(self, queryseq, scorematrix, queryseq_mask=None, scorematrix_mask=None, blank_symbol=None, align='pre'):
        """
        Compute CTC cost, using only the forward pass
        :param queryseq: (L, B)
        :param scorematrix: (T, C+1, B)
        :param queryseq_mask: (L, B)
        :param scorematrix_mask: (T, B)
        :param blank_symbol: scalar, = C by default
        :param align: string, {'pre'/'post'}, indicating how input samples are aligned in one batch
        :return: negative log likelihood averaged over a batch
        """
        if blank_symbol is None:
            # blank_symbol = scorematrix.shape[1] - 1.0
            blank_symbol = tensor.cast(scorematrix.shape[1], floatX) - 1.0
        queryseq_padded, queryseq_mask_padded = self._pad_blanks(queryseq, blank_symbol, queryseq_mask)
        NLL, alphas = self.path_probability(queryseq_padded, scorematrix, queryseq_mask_padded, scorematrix_mask, blank_symbol, align)
        NLL_avg = tensor.mean(NLL)
        return NLL_avg

    @classmethod
    def path_probability(self, queryseq_padded, scorematrix, queryseq_mask_padded=None, scorematrix_mask=None, blank_symbol=None, align='pre'):
        """
        Compute p(l|x) using only the forward variable and log scale
        :param queryseq_padded: (2L+1, B)
        :param scorematrix: (T, C+1, B)
        :param queryseq_mask_padded: (2L+1, B)
        :param scorematrix_mask: (T, B)
        :param blank_symbol: = C by default
        :return:
        """
        if blank_symbol is None:
            # blank_symbol = scorematrix.shape[1] - 1.0
            blank_symbol = tensor.cast(scorematrix.shape[1], floatX) - 1.0
        if queryseq_mask_padded is None:
            queryseq_mask_padded = tensor.ones_like(queryseq_padded, dtype=floatX)
        if scorematrix_mask is None:
            scorematrix_mask = tensor.ones([scorematrix.shape[0], scorematrix.shape[2]])

        pred_y = self._class_batch_to_labeling_batch(queryseq_padded, scorematrix, scorematrix_mask)  # (T, 2L+1, B), reshaped scorematrix
        r2, r3 = self._recurrence_relation(queryseq_padded, queryseq_mask_padded, blank_symbol)       # r2 (2L+1, 2L+1), r3 (2L+1, 2L+1, B)

        def step(p_curr, p_prev):
            p1 = p_prev
            p2 = self._log_dot_matrix(p1, r2)
            p3 = self._log_dot_tensor(p1, r3)
            p123 = self._log_add(p3, self._log_add(p1, p2))
            return p_curr.T + p123 + self._epslog(queryseq_mask_padded.T)

        alphas, _ = theano.scan(
                step,
                sequences=[self._epslog(pred_y)],
                outputs_info=[self._epslog(tensor.eye(queryseq_padded.shape[0])[0] * tensor.ones(queryseq_padded.T.shape))])

        B = alphas.shape[1]
        LL = tensor.sum(queryseq_mask_padded, axis=0, dtype='int32')
        if align == 'pre':
            TL = tensor.sum(scorematrix_mask, axis=0, dtype='int32')
            NLL = -self._log_add(alphas[TL - 1, tensor.arange(B), LL - 1],
                                 alphas[TL - 1, tensor.arange(B), LL - 2])
        else:  # align == 'post'
            NLL = -self._log_add(alphas[-1, tensor.arange(B), LL - 1],
                                 alphas[-1, tensor.arange(B), LL - 2])
        return NLL, alphas

    @staticmethod
    def _epslog(x):
        return tensor.cast(tensor.log(tensor.clip(x, 1E-12, 1E12)),
                           theano.config.floatX)

    @staticmethod
    def _log_add(a, b):
        max_ = tensor.maximum(a, b)
        return max_ + tensor.log1p(tensor.exp(a + b - 2 * max_))

    @staticmethod
    def _log_dot_matrix(x, z):
        inf = 1E12
        log_dot = tensor.dot(x, z)
        zeros_to_minus_inf = (z.max(axis=0) - 1) * inf
        return log_dot + zeros_to_minus_inf

    @staticmethod
    def _log_dot_tensor(x, z):
        inf = 1E12
        log_dot = (x.dimshuffle(1, 'x', 0) * z).sum(axis=0).T
        zeros_to_minus_inf = (z.max(axis=0) - 1) * inf
        return log_dot + zeros_to_minus_inf.T


def ctc_path_probability(scorematrix, queryseq, blank):
    """
    Compute path probability based on CTC algorithm, only forward pass is used.
    Batch not supported, for batch version, refer to the CTC class above
    Speed much slower than the numba & cython version (51.5min vs ~3.9min on word_correction_CTC experiment)
    :param scorematrix: (T, C+1)
    :param queryseq:    (L, 1)
    :param blank:       scalar, blank symbol
    :return: (NLL, alphas), NLL > 0 (smaller is better, = -log(p(l|x)); alphas is the forward variable)
    """

    def update_s(s, alphas, scorematrix, queryseq, blank, t):
        l = (s - 1) // 2
        alphas = ifelse(tensor.eq(s % 2, 0),
                        ifelse(tensor.eq(s, 0),
                               tensor.set_subtensor(alphas[s, t], alphas[s, t - 1] * scorematrix[blank, t]),
                               tensor.set_subtensor(alphas[s, t],
                                                    (alphas[s, t - 1] + alphas[s - 1, t - 1]) * scorematrix[blank, t]),
                               name='for_blank_symbol'),
                        ifelse(tensor.or_(tensor.eq(s, 1), tensor.eq(queryseq[l], queryseq[l - 1])),
                               tensor.set_subtensor(alphas[s, t],
                                                    (alphas[s, t - 1] + alphas[s - 1, t - 1]) * scorematrix[
                                                        queryseq[l], t]),
                               tensor.set_subtensor(alphas[s, t],
                                                    (alphas[s, t - 1] + alphas[s - 1, t - 1] + alphas[s - 2, t - 1]) *
                                                    scorematrix[queryseq[l], t]),
                               name='for_same_label_twice'))
        return alphas

    def update_t(t, LLForward, alphas, scorematrix, queryseq, blank, T, L2):
        start = tensor.max([0, L2 - 2 * (T - t)])
        end = tensor.min([2 * t + 2, L2])
        s = tensor.arange(start, end)
        results, _ = theano.scan(fn=update_s, sequences=[s], non_sequences=[scorematrix, queryseq, blank, t],
                                 outputs_info=[alphas], name='scan_along_s')
        alphas = results[-1]
        c = tensor.sum(alphas[start:end, t])
        c = tensor.max([1e-15, c])
        alphas = tensor.set_subtensor(alphas[start:end, t], alphas[start:end, t] / c)
        LLForward += tensor.log(c)
        return LLForward, alphas

    L = queryseq.shape[0]                                                 # Length of label sequence
    L2 = 2 * L + 1                                                        # Length of label sequence padded with blanks
    T = scorematrix.shape[1]                                              # time length
    alphas = tensor.zeros((L2, T))
    # Initialize alphas and forward pass
    alphas = tensor.set_subtensor(alphas[[0, 1], 0], scorematrix[[blank, queryseq[0]], 0])
    c = tensor.sum(alphas[:, 0])
    alphas = tensor.set_subtensor(alphas[:, 0], alphas[:, 0] / c)
    LLForward = tensor.log(c)
    t = tensor.arange(1, T)
    results, _ = theano.scan(fn=update_t, sequences=[t], non_sequences=[scorematrix, queryseq, blank, T, L2],
                             outputs_info=[LLForward, alphas], name='scan_along_t')
    NLL, alphas = ifelse(tensor.gt(T, 1), (-results[0][-1], results[1][-1]), (-LLForward, alphas))
    return NLL, alphas






# coding:utf-8
'''
  Dandelion objectives pool
  Created   :   9, 20, 2017
  Revised   :   9, 20, 2017
'''

import theano
import theano.tensor as tensor
from .ctc_theano import CTC_Timescale, CTC_Logscale
from .util import as_theano_expression, one_hot

# __all__ = [
#     "categorical_crossentropy",
#     "ctc_cost_logscale",
#     "ctc_cost_timescale",
#     "ctc_best_path_decode",
#     "ctc_CER"
# ]

def align_targets(predictions, targets):
    """Helper function turning a target 1D vector into a column if needed.
    This way, combining a network of a single output unit with a target vector
    works as expected by most users, not broadcasting outputs against targets.

    Parameters
    ----------
    predictions : Theano tensor
        Expression for the predictions of a neural network.
    targets : Theano tensor
        Expression or variable for corresponding targets.

    Returns
    -------
    predictions : Theano tensor
        The predictions unchanged.
    targets : Theano tensor
        If `predictions` is a column vector and `targets` is a 1D vector,
        returns `targets` turned into a column vector. Otherwise, returns
        `targets` unchanged.
    """
    if (getattr(predictions, 'broadcastable', None) == (False, True) and
            getattr(targets, 'ndim', None) == 1):
        targets = as_theano_expression(targets).dimshuffle(0, 'x')
    return predictions, targets


def binary_crossentropy(predictions, targets):
    """Computes the binary cross-entropy between predictions and targets.

    .. math:: L = -t \\log(p) - (1 - t) \\log(1 - p)

    Parameters
    ----------
    predictions : Theano tensor
        Predictions in (0, 1), such as sigmoidal output of a neural network.
    targets : Theano tensor
        Targets in [0, 1], such as ground truth labels.

    Returns
    -------
    Theano tensor
        An expression for the element-wise binary cross-entropy.

    Notes
    -----
    This is the loss function of choice for binary classification problems
    and sigmoid output units.
    """
    predictions, targets = align_targets(predictions, targets)
    return theano.tensor.nnet.binary_crossentropy(predictions, targets)


def categorical_crossentropy(predictions, targets, eps=1e-7):
    """Computes the categorical cross-entropy between predictions and targets.

    .. math:: L_i = - \\sum_j{t_{i,j} \\log(p_{i,j})}

    :math:`p` are the predictions, :math:`t` are the targets, :math:`i`
    denotes the data point and :math:`j` denotes the class.

    Parameters
    ----------
    predictions : Theano 2D tensor
        Predictions in (0, 1), such as softmax output of a neural network,
        with data points in rows and class probabilities in columns.
    targets : Theano 2D tensor or 1D tensor
        Either targets in [0, 1] matching the layout of `predictions`, or
        a vector of int giving the correct class index per data point.
        In the case of an integer vector argument, each element
        represents the position of the '1' in a one-hot encoding.
    eps: epsilon added to `predictions` to prevent numerical unstability when using with softmax activation

    Returns
    -------
    Theano 1D tensor
        An expression for the item-wise categorical cross-entropy.

    Notes
    -----
    This is the loss function of choice for multi-class classification
    problems and softmax output units. For hard targets, i.e., targets
    that assign all of the probability to a single class per data point,
    providing a vector of int for the targets is usually slightly more
    efficient than providing a matrix with a single 1.0 per row.
    """
    if eps > 0:
        predictions = theano.tensor.clip(predictions, eps, 1.0 - eps)
    return theano.tensor.nnet.categorical_crossentropy(predictions, targets)

def categorical_crossentropy_log(log_predictions, targets, m=None):
    """
    Computes the categorical cross-entropy in log-domain.
    :param log_predictions: usually returned by log_softmax()
    :param targets: Theano 2D tensor or 1D tensor
    :param m: possible max value of `targets`'s element
    :return:
    """

    if targets.ndim == log_predictions.ndim:
        return -tensor.sum(targets * log_predictions, axis=log_predictions.ndim-1)
    elif targets.ndim == log_predictions.ndim - 1:
        return -tensor.sum(one_hot(targets, m=m) * log_predictions, axis=log_predictions.ndim-1)
    else:
        raise TypeError('shape mismatch between `log_predictions` and `targets`')

def squared_error(a, b):
    """Computes the element-wise squared difference between two tensors.

    .. math:: L = (p - t)^2

    Parameters
    ----------
    a, b : Theano tensor
        The tensors to compute the squared difference between.

    Returns
    -------
    Theano tensor
        An expression for the element-wise squared difference.

    Notes
    -----
    This is the loss function of choice for many regression problems
    or auto-encoders with linear output units.
    """
    a, b = align_targets(a, b)
    return theano.tensor.square(a - b)


def aggregate(loss, weights=None, mode='mean'):
    """Aggregates an element- or item-wise loss to a scalar loss.

    Parameters
    ----------
    loss : Theano tensor
        The loss expression to aggregate.
    weights : Theano tensor, optional
        The weights for each element or item, must be broadcastable to
        the same shape as `loss` if given. If omitted, all elements will
        be weighted the same.
    mode : {'mean', 'sum', 'normalized_sum'}
        Whether to aggregate by averaging, by summing or by summing and
        dividing by the total weights (which requires `weights` to be given).

    Returns
    -------
    Theano scalar
        A scalar loss expression suitable for differentiation.

    Notes
    -----
    By supplying binary weights (i.e., only using values 0 and 1), this
    function can also be used for masking out particular entries in the
    loss expression. Note that masked entries still need to be valid
    values, not-a-numbers (NaNs) will propagate through.

    When applied to batch-wise loss expressions, setting `mode` to
    ``'normalized_sum'`` ensures that the loss per batch is of a similar
    magnitude, independent of associated weights. However, it means that
    a given data point contributes more to the loss when it shares a batch
    with low-weighted or masked data points than with high-weighted ones.
    """
    if weights is not None:
        loss = loss * weights
    if mode == 'mean':
        return loss.mean()
    elif mode == 'sum':
        return loss.sum()
    elif mode == 'normalized_sum':
        if weights is None:
            raise ValueError("require weights for mode='normalized_sum'")
        return loss.sum() / weights.sum()
    else:
        raise ValueError("mode must be 'mean', 'sum' or 'normalized_sum', "
                         "got %r" % mode)


def binary_hinge_loss(predictions, targets, delta=1, log_odds=None,
                      binary=True):
    """Computes the binary hinge loss between predictions and targets.

    .. math:: L_i = \\max(0, \\delta - t_i p_i)

    Parameters
    ----------
    predictions : Theano tensor
        Predictions in (0, 1), such as sigmoidal output of a neural network
        (or log-odds of predictions depending on `log_odds`).
    targets : Theano tensor
        Targets in {0, 1} (or in {-1, 1} depending on `binary`), such as
        ground truth labels.
    delta : scalar, default 1
        The hinge loss margin
    log_odds : bool, default None
        ``False`` if predictions are sigmoid outputs in (0, 1), ``True`` if
        predictions are sigmoid inputs, or log-odds. If ``None``, will assume
        ``True``, but warn that the default will change to ``False``.
    binary : bool, default True
        ``True`` if targets are in {0, 1}, ``False`` if they are in {-1, 1}

    Returns
    -------
    Theano tensor
        An expression for the element-wise binary hinge loss

    Notes
    -----
    This is an alternative to the binary cross-entropy loss for binary
    classification problems.

    Note that it is a drop-in replacement only when giving ``log_odds=False``.
    Otherwise, it requires log-odds rather than sigmoid outputs. Be aware that
    depending on the Theano version, ``log_odds=False`` with a sigmoid
    output layer may be less stable than ``log_odds=True`` with a linear layer.
    """
    if log_odds is None:  # pragma: no cover
        log_odds = True
        raise FutureWarning(
                "The `log_odds` argument to `binary_hinge_loss` will change "
                "its default to `False` in a future version. Explicitly give "
                "`log_odds=True` to retain current behavior in your code, "
                "but also check the documentation if this is what you want.")
    if not log_odds:
        predictions = theano.tensor.log(predictions / (1 - predictions))
    if binary:
        targets = 2 * targets - 1
    predictions, targets = align_targets(predictions, targets)
    return theano.tensor.nnet.relu(delta - predictions * targets)


def multiclass_hinge_loss(predictions, targets, delta=1):
    """Computes the multi-class hinge loss between predictions and targets.

    .. math:: L_i = \\max_{j \\not = t_i} (0, p_j - p_{t_i} + \\delta)

    Parameters
    ----------
    predictions : Theano 2D tensor
        Predictions in (0, 1), such as softmax output of a neural network,
        with data points in rows and class probabilities in columns.
    targets : Theano 2D tensor or 1D tensor
        Either a vector of int giving the correct class index per data point
        or a 2D tensor of one-hot encoding of the correct class in the same
        layout as predictions (non-binary targets in [0, 1] do not work!)
    delta : scalar, default 1
        The hinge loss margin

    Returns
    -------
    Theano 1D tensor
        An expression for the item-wise multi-class hinge loss

    Notes
    -----
    This is an alternative to the categorical cross-entropy loss for
    multi-class classification problems
    """
    num_cls = predictions.shape[1]
    if targets.ndim == predictions.ndim - 1:
        targets = theano.tensor.extra_ops.to_one_hot(targets, num_cls)
    elif targets.ndim != predictions.ndim:
        raise TypeError('rank mismatch between targets and predictions')
    corrects = predictions[targets.nonzero()]
    rest = theano.tensor.reshape(predictions[(1-targets).nonzero()],
                                 (-1, num_cls-1))
    rest = theano.tensor.max(rest, axis=1)
    return theano.tensor.nnet.relu(rest - corrects + delta)


def binary_accuracy(predictions, targets, threshold=0.5):
    """Computes the binary accuracy between predictions and targets.

    .. math:: L_i = \\mathbb{I}(t_i = \mathbb{I}(p_i \\ge \\alpha))

    Parameters
    ----------
    predictions : Theano tensor
        Predictions in [0, 1], such as a sigmoidal output of a neural network,
        giving the probability of the positive class
    targets : Theano tensor
        Targets in {0, 1}, such as ground truth labels.
    threshold : scalar, default: 0.5
        Specifies at what threshold to consider the predictions being of the
        positive class

    Returns
    -------
    Theano tensor
        An expression for the element-wise binary accuracy in {0, 1}

    Notes
    -----
    This objective function should not be used with a gradient calculation;
    its gradient is zero everywhere. It is intended as a convenience for
    validation and testing, not training.

    To obtain the average accuracy, call :func:`theano.tensor.mean()` on the
    result, passing ``dtype=theano.config.floatX`` to compute the mean on GPU.
    """
    predictions, targets = align_targets(predictions, targets)
    predictions = theano.tensor.ge(predictions, threshold)
    return theano.tensor.eq(predictions, targets)


def categorical_accuracy(predictions, targets, top_k=1):
    """Computes the categorical accuracy between predictions and targets.

    .. math:: L_i = \\mathbb{I}(t_i = \\operatorname{argmax}_c p_{i,c})

    Can be relaxed to allow matches among the top :math:`k` predictions:

    .. math::
        L_i = \\mathbb{I}(t_i \\in \\operatorname{argsort}_c (-p_{i,c})_{:k})

    Parameters
    ----------
    predictions : Theano 2D tensor
        Predictions in (0, 1), such as softmax output of a neural network,
        with data points in rows and class probabilities in columns.
    targets : Theano 2D tensor or 1D tensor
        Either a vector of int giving the correct class index per data point
        or a 2D tensor of 1 hot encoding of the correct class in the same
        layout as predictions
    top_k : int
        Regard a prediction to be correct if the target class is among the
        `top_k` largest class probabilities. For the default value of 1, a
        prediction is correct only if the target class is the most probable.

    Returns
    -------
    Theano 1D tensor
        An expression for the item-wise categorical accuracy in {0, 1}

    Notes
    -----
    This is a strictly non differential function as it includes an argmax.
    This objective function should never be used with a gradient calculation.
    It is intended as a convenience for validation and testing not training.

    To obtain the average accuracy, call :func:`theano.tensor.mean()` on the
    result, passing ``dtype=theano.config.floatX`` to compute the mean on GPU.
    """
    if targets.ndim == predictions.ndim:
        targets = theano.tensor.argmax(targets, axis=-1)
    elif targets.ndim != predictions.ndim - 1:
        raise TypeError('rank mismatch between targets and predictions')

    if top_k == 1:
        # standard categorical accuracy
        top = theano.tensor.argmax(predictions, axis=-1)
        return theano.tensor.eq(top, targets)
    else:
        # top-k accuracy
        top = theano.tensor.argsort(predictions, axis=-1)
        # (Theano cannot index with [..., -top_k:], we need to simulate that)
        top = top[[slice(None) for _ in range(top.ndim - 1)] +
                  [slice(-top_k, None)]]
        targets = theano.tensor.shape_padaxis(targets, axis=-1)
        return theano.tensor.any(theano.tensor.eq(top, targets), axis=-1)


def ctc_cost_timescale(seq, sm, seq_mask=None, sm_mask=None, blank_symbol=None):
    """
    seq (B, L), sm (B, T, C+1), seq_mask (B, L), sm_mask (B, T)
    Compute CTC cost, using only the forward pass
    :param seq: query sequence, (L, B), float32
    :param sm: score matrix, (T, C+1, B), float32
    :param seq_mask: mask for query sequence, (L, B), float32
    :param sm_mask: mask for score matrix, (T, B), float32
    :param blank_symbol: scalar
    :return: negative log likelihood averaged over a batch
    """
    queryseq = seq.T
    scorematrix = sm.dimshuffle(1, 2, 0)
    if seq_mask is None:
        queryseq_mask = None
    else:
        queryseq_mask = seq_mask.T
    if sm_mask is None:
        scorematrix_mask = None
    else:
        scorematrix_mask = sm_mask.T

    return CTC_Timescale.cost(queryseq, scorematrix, queryseq_mask, scorematrix_mask, blank_symbol)

def ctc_cost_logscale(seq, sm, seq_mask=None, sm_mask=None, blank_symbol=None, align='pre'):
    """
    Compute CTC cost, using only the forward pass
    :param seq: query sequence, (L, B), float32
    :param sm: score matrix, (T, C+1, B), float32
    :param seq_mask: mask for query sequence, (L, B), float32
    :param sm_mask: mask for score matrix, (T, B), float32
    :param blank_symbol: scalar
    :return: negative log likelihood averaged over a batch
    """
    queryseq = tensor.addbroadcast(seq.T)
    scorematrix = sm.dimshuffle(1, 2, 0)
    if seq_mask is None:
        queryseq_mask = None
    else:
        queryseq_mask = seq_mask.T
    if sm_mask is None:
        scorematrix_mask = None
    else:
        scorematrix_mask = sm_mask.T

    return CTC_Logscale.cost(queryseq, scorematrix, queryseq_mask, scorematrix_mask, blank_symbol, align)

def ctc_best_path_decode(Y, Y_mask=None, blank_symbol=None):
    """
    Decode the network output scorematrix by best-path-decoding scheme
    :param Y: output of a network, with shape (batch, timesteps, Nclass+1)
    :param Y_mask: mask of Y, with shape (batch, timesteps)
    :return:
    """
    scorematrix = Y.dimshuffle(1, 2, 0)
    if Y_mask is None:
        scorematrix_mask = None
    else:
        scorematrix_mask = Y_mask.dimshuffle(1, 0)
    if blank_symbol is None:
        blank_symbol = Y.shape[2] - 1
    resultseq, resultseq_mask = CTC_Logscale.best_path_decode(scorematrix, scorematrix_mask, blank_symbol)
    return resultseq, resultseq_mask

def ctc_CER(resultseq, targetseq, resultseq_mask=None, targetseq_mask=None):
    return CTC_Logscale.calc_CER(resultseq, targetseq, resultseq_mask, targetseq_mask)



# coding:utf-8
# Test for Conv2D class
# Created   :   1, 31, 2018
# Revised   :   1, 31, 2018
# All rights reserved
#------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'
import os, sys, psutil
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"

import theano
from theano import tensor
from dandelion.module import *
from dandelion.activation import *
from lasagne.layers import InputLayer, Conv2DLayer, get_output
import lasagne.nonlinearities as LACT
import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)

class build_model_D(Module):
    def __init__(self, in_channel=3, out_channel=3, kernel_size=(3,3), stride=(1,1), pad='valid', dilation=(1,1), num_groups=1):
        super().__init__()
        self.conv2d = Conv2D(in_channels=in_channel, out_channels=out_channel, kernel_size=kernel_size, stride=stride,
                             pad=pad, dilation=dilation, num_groups=num_groups)
        self.predict = self.forward

    def forward(self, x):
        """

        :param x: (B, C, H, W)
        :return:
        """
        x = self.conv2d.forward(x)
        x = relu(x)
        return x

def build_model_L(in_channel=3, out_channel=3, kernel_size=(3,3), stride=(1,1), pad='valid', dilation=(1,1), num_groups=1):
    input_var = tensor.ftensor4('x')  # (B, C, H, W)
    input0 = InputLayer(shape=(None, in_channel, None, None), input_var=input_var, name='input0')
    conv0  = Conv2DLayer(input0, num_filters=out_channel, filter_size=kernel_size, stride=stride, pad=pad, nonlinearity=LACT.rectify,
                        name='conv0')
    return conv0


if __name__ == '__main__':
    import numpy as np
    from lasagne_ext.utils import get_layer_by_name

    in_channel = 3; out_channel = 3;kernel_size = (3, 3); stride = (1, 1); pad = 'valid';dilation = (1,1);num_groups = 1
    model_D = build_model_D(in_channel=in_channel, out_channel=out_channel, kernel_size=kernel_size, stride=stride,
                             pad=pad, dilation=dilation, num_groups=num_groups)
    model_L = build_model_L(in_channel=in_channel, out_channel=out_channel, kernel_size=kernel_size, stride=stride,
                             pad=pad)

    W = np.random.rand(out_channel, in_channel, kernel_size[0], kernel_size[1]).astype(np.float32)
    b = np.random.rand(out_channel).astype(np.float32)

    model_D.conv2d.W.set_value(W)
    model_D.conv2d.b.set_value(b)

    conv_L = get_layer_by_name(model_L, 'conv0')
    conv_L.W.set_value(W)
    conv_L.b.set_value(b)

    X = get_layer_by_name(model_L, 'input0').input_var
    y_D = model_D.forward(X)
    y_L = get_output(model_L)

    fn_D = theano.function([X], y_D, no_default_updates=True, on_unused_input='ignore')
    fn_L = theano.function([X], y_L, no_default_updates=True, on_unused_input='ignore')

    for i in range(20):
        x = np.random.rand(3, in_channel, 32, 32).astype(np.float32) - 0.5
        y_D = fn_D(x)
        y_L = fn_L(x)
        diff = np.max(np.abs(y_D - y_L))
        print('i=%d, diff=%0.6f' % (i, diff))
        if diff>1e-4:
            print('y_D=\n', y_D)
            print('y_L=\n', y_L)
            raise ValueError('diff is too big')

    print('Test passed')






# coding:utf-8
'''
Model definition of ancient VGG CNN nets
Created   :   7,  6, 2018
Revised   :   7,  6, 2018
All rights reserved
'''
# ------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'

import theano.tensor as tensor
from ..module import *
from ..functional import *
from ..activation import *


class model_VGG16(Module):
    """
    VGG16-net reference implementation: 13 cnn + 3 dense + 2 dropout.
    """

    def __init__(self, channel=3, im_height=224, im_width=224, Nclass=1000, kernel_size=3, border_mode=(1, 1)):
        super().__init__()
        self.conv1_1 = Conv2D(in_channels=channel, out_channels=64, kernel_size=kernel_size, pad=border_mode, input_shape=(im_height, im_width))
        self.conv1_2 = Conv2D(in_channels=64,  out_channels=64,  kernel_size=kernel_size, pad=border_mode)
        self.conv2_1 = Conv2D(in_channels=64,  out_channels=128, kernel_size=kernel_size, pad=border_mode)
        self.conv2_2 = Conv2D(in_channels=128, out_channels=128, kernel_size=kernel_size, pad=border_mode)
        self.conv3_1 = Conv2D(in_channels=128, out_channels=256, kernel_size=kernel_size, pad=border_mode)
        self.conv3_2 = Conv2D(in_channels=256, out_channels=256, kernel_size=kernel_size, pad=border_mode)
        self.conv3_3 = Conv2D(in_channels=256, out_channels=256, kernel_size=kernel_size, pad=border_mode)
        self.conv4_1 = Conv2D(in_channels=256, out_channels=512, kernel_size=kernel_size, pad=border_mode)
        self.conv4_2 = Conv2D(in_channels=512, out_channels=512, kernel_size=kernel_size, pad=border_mode)
        self.conv4_3 = Conv2D(in_channels=512, out_channels=512, kernel_size=kernel_size, pad=border_mode)
        self.conv5_1 = Conv2D(in_channels=512, out_channels=512, kernel_size=kernel_size, pad=border_mode)
        self.conv5_2 = Conv2D(in_channels=512, out_channels=512, kernel_size=kernel_size, pad=border_mode)
        self.conv5_3 = Conv2D(in_channels=512, out_channels=512, kernel_size=kernel_size, pad=border_mode)
        self.fc6     = Dense(input_dims=512 * im_height//32 * im_width//32, output_dim=4096)
        self.fc7     = Dense(input_dims=4096, output_dim=4096)
        self.fc8     = Dense(input_dims=4096, output_dim=Nclass)
        self.fc6_dropout = Dropout()
        self.fc7_dropout = Dropout()


    def forward(self, x):
        """
        :param x: (B, C, H, W)
        :return:
        """
        self.work_mode = 'train'

        x = relu(self.conv1_1.forward(x))  # (B, 64, 224, 224)
        x = relu(self.conv1_2.forward(x))
        x = pool_2d(x, ws=(2, 2))           # (B, 64, 112, 112)
        x = relu(self.conv2_1.forward(x))
        x = relu(self.conv2_2.forward(x))
        x = pool_2d(x, ws=(2, 2))          # (B, 128, 56, 56)
        x = relu(self.conv3_1.forward(x))
        x = relu(self.conv3_2.forward(x))
        x = relu(self.conv3_3.forward(x))
        x = pool_2d(x, ws=(2, 2))          # (B, 256, 28, 28)
        x = relu(self.conv4_1.forward(x))
        x = relu(self.conv4_2.forward(x))
        x = relu(self.conv4_3.forward(x))
        x = pool_2d(x, ws=(2, 2))          # (B, 512, 14, 14)
        x = relu(self.conv5_1.forward(x))
        x = relu(self.conv5_2.forward(x))
        x = relu(self.conv5_3.forward(x))
        x = pool_2d(x, ws=(2, 2))          # (B, 512, 7, 7)
        x = tensor.flatten(x, ndim=2)  # (B, 512 * 7 * 7)
        x = relu(self.fc6.forward(x))
        x = self.fc6_dropout.forward(x, p=0.5)
        x = relu(self.fc7.forward(x))
        x = self.fc7_dropout.forward(x, p=0.5)
        x = softmax(self.fc8.forward(x))

        return x

    def predict(self, x):
        self.work_mode = 'inference'

        x = relu(self.conv1_1.predict(x))  # (B, 64, 224, 224)
        x = relu(self.conv1_2.predict(x))
        x = pool_2d(x, ws=(2, 2))           # (B, 64, 112, 112)
        x = relu(self.conv2_1.predict(x))
        x = relu(self.conv2_2.predict(x))
        x = pool_2d(x, ws=(2, 2))          # (B, 128, 56, 56)
        x = relu(self.conv3_1.predict(x))
        x = relu(self.conv3_2.predict(x))
        x = relu(self.conv3_3.predict(x))
        x = pool_2d(x, ws=(2, 2))          # (B, 256, 28, 28)
        x = relu(self.conv4_1.predict(x))
        x = relu(self.conv4_2.predict(x))
        x = relu(self.conv4_3.predict(x))
        x = pool_2d(x, ws=(2, 2))          # (B, 512, 14, 14)
        x = relu(self.conv5_1.predict(x))
        x = relu(self.conv5_2.predict(x))
        x = relu(self.conv5_3.predict(x))
        x = pool_2d(x, ws=(2, 2))          # (B, 512, 7, 7)
        x = tensor.flatten(x, ndim=2)      # (B, 512 * 7 * 7)
        x = relu(self.fc6.predict(x))
        x = relu(self.fc7.predict(x))
        x = softmax(self.fc8.predict(x))

        return x



# coding:utf-8
# Test for categorical_crossentropy_log
# Created   :   7,  3, 2018
# Revised   :   7,  3, 2018
# All rights reserved
#------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'
import os, numpy as np
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"

import theano
from theano import tensor
from dandelion.module import *
from dandelion.activation import softmax, log_softmax
from dandelion.objective import categorical_crossentropy, categorical_crossentropy_log
from dandelion.util import one_hot

import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)

if __name__ == '__main__':

    if 0:
        x = tensor.fmatrix()  # (B, N)
        # y = tensor.ivector()  # (B,)
        y = tensor.fmatrix()  # (B, N)

        x1 = softmax(x) # (B, N)
        r1 = categorical_crossentropy(x1, y, eps=0.0)

        x2 = log_softmax(x)
        r2 = categorical_crossentropy_log(x2, y)

        f1 = theano.function([x,y], r1)
        f2 = theano.function([x,y], r2)

        for i in range(100):
            B, N = np.random.randint(low=1, high=32), np.random.randint(low=2, high=100)
            X = np.random.rand(B, N)
            Y = np.random.randint(low=0, high=N, size=(B,))
            Y = np.eye(N)[Y]


            X = X.astype('float32')
            Y = Y.astype('float32')

            r1 = f1(X, Y)
            r2 = f2(X, Y)
            dif = np.sum(np.abs(r1-r2))
            if dif >1e-7:
                print(r1)
                print(r2)
                raise ValueError('r1 != r2')
    if 1:
        N = np.random.randint(low=2, high=100)
        x = tensor.fmatrix()  # (B, N)
        y = tensor.ivector()  # (B,)

        x1 = softmax(x) # (B, N)
        r1 = categorical_crossentropy(x1, y, eps=0.0)

        x2 = log_softmax(x)
        r2 = categorical_crossentropy_log(x2, y, m=N)

        f1 = theano.function([x,y], r1)
        f2 = theano.function([x,y], r2)

        for i in range(100):
            # B, N = np.random.randint(low=1, high=32), np.random.randint(low=2, high=100)
            B = np.random.randint(low=1, high=32)
            X = np.random.rand(B, N)
            Y = np.random.randint(low=0, high=N, size=(B,))

            X = X.astype('float32')
            Y = Y.astype('int32')

            r1 = f1(X, Y)
            # print('r1=', r1)
            r2 = f2(X, Y)
            # print('r2=', r2)
            dif = np.max(np.abs(r1-r2))
            if dif >1e-6:
                print('r1=',r1)
                print('r2=',r2)
                raise ValueError('r1 != r2')

    print('Test passed')






# coding:utf-8
# Unit test for LSTM class
# Created   :   1, 30, 2018
# Revised   :   1, 30, 2018
# All rights reserved
#------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'
import os, sys, psutil
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"

import theano
from theano import tensor
from dandelion.module import *
from dandelion.activation import *
from lasagne.layers import InputLayer, DenseLayer, LSTMLayer, get_output, Upscale2DLayer, TransposedConv2DLayer
import lasagne.nonlinearities as LACT

class build_model_D(Module):
    def __init__(self, in_dim=3, out_dim=3):
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.lstm = LSTM(input_dims=self.in_dim, hidden_dim=self.out_dim, peephole=True)
        self.predict = self.forward

    def forward(self, x):
        """

        :param x: (B, T, D)
        :return:
        """
        x = x.dimshuffle((1, 0, 2)) # ->(T, B, D)
        x = self.lstm.forward(x, backward=True, only_return_final=True)
        # x = x.dimshuffle((1, 0, 2)) # ->(B, T, D)
        # x = tanh(x)
        return x

def build_model_L(in_dim=3, out_dim=3):
    input_var = tensor.ftensor3('x')  # (B, T, D)
    input0 = InputLayer(shape=(None, None, in_dim), input_var=input_var, name='input0')
    lstm0 = LSTMLayer(input0, num_units=out_dim, precompute_input=True, nonlinearity=LACT.tanh,
                      backwards=True, only_return_final=True, learn_init=True, consume_less='None',
                      name='lstm0')
    return lstm0


if __name__ == '__main__':
    import numpy as np
    from lasagne_ext.utils import get_layer_by_name

    in_dim, out_dim = 32, 3
    model_D = build_model_D(in_dim=in_dim, out_dim=out_dim)
    model_L = build_model_L(in_dim=in_dim, out_dim=out_dim)

    W_in = np.random.rand(in_dim, 4*out_dim).astype(np.float32)
    b_in = np.random.rand(4*out_dim).astype(np.float32)
    W_hid = np.random.rand(out_dim, 4*out_dim).astype(np.float32)
    h_ini = np.random.rand(out_dim).astype(np.float32)
    c_ini = np.random.rand(out_dim).astype(np.float32)
    w_cell_to_igate = np.random.rand(out_dim).astype(np.float32)
    w_cell_to_fgate = np.random.rand(out_dim).astype(np.float32)
    w_cell_to_ogate = np.random.rand(out_dim).astype(np.float32)

    model_D.lstm.W_in.set_value(W_in)
    model_D.lstm.b_in.set_value(b_in)
    model_D.lstm.W_hid.set_value(W_hid)
    model_D.lstm.h_ini.set_value(h_ini)
    model_D.lstm.c_ini.set_value(c_ini)
    model_D.lstm.w_cell_to_igate.set_value(w_cell_to_igate)
    model_D.lstm.w_cell_to_fgate.set_value(w_cell_to_fgate)
    model_D.lstm.w_cell_to_ogate.set_value(w_cell_to_ogate)

    lstm_L = get_layer_by_name(model_L, 'lstm0')
    lstm_L.W_in_to_ingate.set_value(W_in[:, :out_dim])
    lstm_L.W_in_to_forgetgate.set_value(W_in[:, out_dim:2*out_dim])
    lstm_L.W_in_to_cell.set_value(W_in[:, 2*out_dim:3*out_dim])
    lstm_L.W_in_to_outgate.set_value(W_in[:, 3*out_dim:])

    lstm_L.W_hid_to_ingate.set_value(W_hid[:, :out_dim])
    lstm_L.W_hid_to_forgetgate.set_value(W_hid[:, out_dim:2*out_dim])
    lstm_L.W_hid_to_cell.set_value(W_hid[:, 2*out_dim:3*out_dim])
    lstm_L.W_hid_to_outgate.set_value(W_hid[:, 3*out_dim:])

    lstm_L.b_ingate.set_value(b_in[:out_dim])
    lstm_L.b_forgetgate.set_value(b_in[out_dim:2*out_dim])
    lstm_L.b_cell.set_value(b_in[2*out_dim:3*out_dim])
    lstm_L.b_outgate.set_value(b_in[3*out_dim:])

    lstm_L.hid_init.set_value(h_ini.reshape((1, out_dim)))
    lstm_L.cell_init.set_value(c_ini.reshape((1, out_dim)))

    lstm_L.W_cell_to_ingate.set_value(w_cell_to_igate)
    lstm_L.W_cell_to_forgetgate.set_value(w_cell_to_fgate)
    lstm_L.W_cell_to_outgate.set_value(w_cell_to_ogate)

    X = get_layer_by_name(model_L, 'input0').input_var
    y_D = model_D.forward(X)
    y_L = get_output(model_L)

    fn_D = theano.function([X], y_D, no_default_updates=True, on_unused_input='ignore')
    fn_L = theano.function([X], y_L, no_default_updates=True, on_unused_input='ignore')

    for i in range(20):
        x = np.random.rand(4, 16, in_dim).astype(np.float32)
        y_D = fn_D(x)
        y_L = fn_L(x)
        diff = np.max(np.abs(y_D - y_L))
        print('i=%d, diff=%0.6f' % (i, diff))
        if diff>1e-4:
            print('y_D=\n', y_D)
            print('y_L=\n', y_L)
            raise ValueError('diff is too big')

    print('Test passed')






# coding:utf-8
# Test for unet, partial.
# Created   :   5, 25, 2018
# Revised   :   5, 25, 2018
# All rights reserved
#------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'
import os, sys, psutil
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"

import theano
from theano import tensor
from dandelion.module import *
from dandelion.activation import *
from dandelion.model.unet import model_Unet

import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)

if __name__ == '__main__':
    im_height, im_width = 65, 63
    model = model_Unet(im_height=im_height, im_width=im_width)
    x = tensor.ftensor4('x')
    y = model.forward(x)
    print('compiling fn...')
    fn = theano.function([x], y, no_default_updates=False)
    print('run fn...')
    input = np.random.rand(7, 1, im_height, im_width).astype(np.float32)
    output = fn(input)
    print(output)
    print(output.shape)

    print('Test passed')






# coding:utf-8
'''
Reference implementation of [Shuffle-net](https://arxiv.org/abs/1707.01083)
Created   :   7,  9, 2018
Revised   :   7,  9, 2018
All rights reserved
'''
# ------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'

import theano.tensor as tensor
from ..module import *
from ..functional import *
from ..activation import *


class DSConv2D(Module):
    """
    Depthwise Separable Convolution (https://arxiv.org/abs/1610.02357)
    """
    def __init__(self, in_channels, out_channels, kernel_size=(3,3), stride=(1,1), dilation=(1,1), pad='valid'):
        """
        depthwise separable convolution = separate 3*3 conv on each input channel, then use 1*1 conv to map to #out_channels output channels
        :param in_channels:
        :param out_channels:
        :param kernel_size:
        :param stride:
        :param dilation:
        :param pad:
        """
        super().__init__()
        self.conv_depthwise = Conv2D(in_channels=in_channels, out_channels=in_channels, kernel_size=kernel_size, stride=stride, pad=pad, dilation=dilation, num_groups=in_channels)
        self.conv_pointwise = Conv2D(in_channels=in_channels, out_channels=out_channels, kernel_size=(1,1), pad='same')

    def forward(self, x):
        """
        :param x: (B, C, H, W)
        :return:
        """
        self.work_mode = 'train'

        x = self.conv_depthwise.forward(x)   # W = (g, 1, h, w), g = group number = in_channels, (h, w) = kernel_size
        x = self.conv_pointwise.forward(x)   # W = (out_channels, n_channels, 1, 1)
        return x

    def predict(self, x):
        self.work_mode = 'inference'

        x = self.conv_depthwise.predict(x)   # W = (g, h, w), g = group number = in_channels, (h, w) = kernel_size
        x = self.conv_pointwise.predict(x)   # W = (in_channels, out_channels, 1, 1)
        return x


class ShuffleUnit(Module):
    """
    Shuffle unit reference implementation (https://arxiv.org/abs/1610.02357)
    """
    def __init__(self, outer_channel=256, inner_channel=64, group_num=4, border_mode='same', batchnorm_mode=1, activation=relu):
        """

        :param outer_channel: channel number of block input
        :param inner_channel: channel number inside the block
        :param group_num: number of convolution groups
        :param border_mode:
        :param batchnorm_mode: {0 | 1 | 2}. 0 means no batch normalization applied; 1 means batch normalization applied to each cnn;
                               2 means batch normalization only applied to the last cnn
        :param activation: default = relu. Note no activation applied to the last element-wise sum output.
        """
        super().__init__()
        self.activation = activation
        self.batchnorm_mode = batchnorm_mode
        self.group_num = group_num
        self.conv1 = Conv2D(in_channels=outer_channel, out_channels=inner_channel, kernel_size=1, pad=border_mode, num_groups=group_num)
        self.conv2 = DSConv2D(in_channels=inner_channel, out_channels=inner_channel, kernel_size=3, pad=border_mode)
        self.conv3 = Conv2D(in_channels=inner_channel, out_channels=outer_channel, kernel_size=1, pad=border_mode)
        if batchnorm_mode == 0:   # no batch normalization
            pass
        elif batchnorm_mode == 1: # batch normalization per convolution
            self.bn1 = BatchNorm(input_shape=(None, inner_channel, None, None))
            self.bn2 = BatchNorm(input_shape=(None, inner_channel, None, None))
            self.bn3 = BatchNorm(input_shape=(None, outer_channel, None, None))
        elif batchnorm_mode == 2: # only one batch normalization at the end
            self.bn3  = BatchNorm(input_shape=(None, outer_channel, None, None))
        else:
            raise ValueError('batchnorm_mode should be {0 | 1 | 2}')

    @staticmethod
    def channel_shuffle(x, group_num):
        """
        Pseudo shuffle channel by dimshuffle & reshape
        :param x: (B, C, H, W)
        :param group_num: int
        :return: 
        """
        x = x.dimshuffle((0, 2, 3, 1)) # (B, H, W, c*g)
        B, H, W, C = x.shape
        x = tensor.reshape(x, (B, H, W, C//group_num, group_num))
        x = x.dimshuffle((0, 1, 2, 4, 3))
        x = x.flatten(ndim=4)
        x = x.dimshuffle(0, 3, 1, 2)
        return x

    def forward(self, x):
        """
        :param x: (B, C, H, W)
        :return:
        """
        self.work_mode = 'train'

        x0 = x
        x  = self.conv1.forward(x)
        if self.batchnorm_mode == 1:
            x = self.bn1.forward(x)
        x = self.activation(x)          # (B, c*g, H, W)
        x = self.channel_shuffle(x, self.group_num)
        x  = self.conv2.forward(x)
        if self.batchnorm_mode == 1:
            x = self.bn2.forward(x)
        x = self.activation(x)
        x  = self.conv3.forward(x)
        if self.batchnorm_mode in {1, 2}:
            x = self.bn3.forward(x)
        x = self.activation(x)
        x = x + x0
        return x

    def predict(self, x):
        self.work_mode = 'inference'

        x0 = x
        x  = self.conv1.predict(x)
        if self.batchnorm_mode == 1:
            x = self.bn1.predict(x)
        x = self.activation(x)          # (B, c*g, H, W)
        x = self.channel_shuffle(x, self.group_num)
        x  = self.conv2.predict(x)
        if self.batchnorm_mode == 1:
            x = self.bn2.predict(x)
        x = self.activation(x)
        x  = self.conv3.predict(x)
        if self.batchnorm_mode in {1, 2}:
            x = self.bn3.predict(x)
        x = self.activation(x)
        x = x + x0
        return x



# coding:utf-8
# Test for upsample_2d
# Created   :   7,  5, 2018
# Revised   :   7,  5, 2018
# All rights reserved
#------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'
import os, sys, psutil
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"

import theano
from theano import tensor
from dandelion.module import *
from dandelion.functional import upsample_2d, upsample_2d_bilinear
from lasagne.layers import InputLayer, get_output, Upscale2DLayer
import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)

class build_model_D(Module):
    def __init__(self, ratio=[2, 3], mode='repeat'):
        super().__init__()
        self.ratio = ratio
        self.mode = mode
        self.predict = self.forward

    def forward(self, x):
        """

        :param x: (B, C, H, W)
        :return:
        """
        x = upsample_2d(x, ratio=self.ratio, mode=self.mode)
        # x = relu(x)
        return x

def build_model_L(ratio=[2,3], mode='repeat'):
    input_var = tensor.ftensor4('x')  # (B, C, H, W)
    input0 = InputLayer(shape=(None, None, None, None), input_var=input_var, name='input0')
    x  = Upscale2DLayer(input0, scale_factor=ratio, mode=mode)
    return x


if __name__ == '__main__':
    import numpy as np
    from lasagne_ext.utils import get_layer_by_name

    ratio = [1, 2]
    mode = 'dilate'

    model_D = build_model_D(ratio=ratio, mode=mode)
    model_L = build_model_L(ratio=ratio, mode=mode)


    X = get_layer_by_name(model_L, 'input0').input_var
    y_D = model_D.forward(X)
    y_L = get_output(model_L)

    fn_D = theano.function([X], y_D, no_default_updates=True, on_unused_input='ignore')
    fn_L = theano.function([X], y_L, no_default_updates=True, on_unused_input='ignore')

    for i in range(20):
        B = np.random.randint(low=1, high=16)
        C = np.random.randint(low=1, high=32)
        H = np.random.randint(low=5, high=256)
        W = np.random.randint(low=5, high=255)
        x = np.random.rand(B, C, H, W).astype(np.float32) - 0.5
        y_D = fn_D(x)
        y_L = fn_L(x)
        # print(y_D)
        diff = np.max(np.abs(y_D - y_L))
        print('i=%d, diff=%0.6f' % (i, diff))
        if diff>1e-4:
            print('y_D=\n', y_D)
            print('y_L=\n', y_L)
            raise ValueError('diff is too big')

    print('Test passed')



